library ormicida.schema.inspector;

import 'dart:mirrors';
import 'dart:typed_data';
import 'package:ormicida_core/ormicida_core.dart';
import 'package:ormicida_bytes/ormicida_bytes.dart';
import 'package:ormicida_codec/ormicida_codec.dart';
import 'package:ormicida_schema/ormicida_schema.dart';

final Inspector inspector = new Inspector();

class Inspector implements InspectableSink {
  final List<Object>  _seen = [];
  CodecListener      _listener;
  bool                _omitDefaultValues;
  
  Inspector();
  
  void clear() {
    _seen.clear();
    _listener = null;
  }
  
  void addProperty(final String name, final value) {
    final listener = _listener;
    listener.onPropertyStart(name);
    _inspect(value, listener);
    listener.onPropertyEnd();
  }
  
  void _pushSeen(object) {
    if (_seen.contains(object))
      throw new CyclicError(object);
    _seen.add(object);
  }
  
  void _popSeen() {
    _seen.removeLast();
  }
  
  void inspect(object, CodecListener listener, Schema schema, {
        bool omitDefaultValues: true}) {
    clear();
    _listener = listener;
    _omitDefaultValues = omitDefaultValues;
    _inspect(object, schema);
  }
  
  void _inspect(final object, final Schema schema) {
    if (object == null) {
      _listener.onValue(null);
      
    } else if (schema == null || schema is IsDynamic) {
      if (object is Map) {
        _pushSeen(object);
        _listener.onMapStart();
        object.forEach((String key, value) {
          _listener.onPropertyStart(key);
          _inspect(value, null);
          _listener.onPropertyEnd();
        });
        _listener.onMapEnd();
        _popSeen();
        
      } else if (object is List && object is! TypedData) {
        _pushSeen(object);
        _listener.onListStart();
        final length = object.length;
        for (var index = 0; index < length; index++) {
          _inspect(object[index], null);
          _listener.onListElement();
        }
        _listener.onListEnd();
        _popSeen();
        
      } else if (object is Inspectable) {
        _pushSeen(object);
        _listener.onMapStart();
        object.inspect(this);
        _listener.onMapEnd();
        _popSeen();
        
      } else {
        _listener.onValue(object);
      }
      
    } else if (schema is IsObject) {
      final properties = schema.properties;
      final length = properties.length;
      _pushSeen(object);
      _listener.onMapStart();
      
      if (object is Map) {
        for (int index = 0; index < length; index++) {
          final property = properties[index];
          final value = object[property.name];
          
          if (! _omitDefaultValues || value != property.defaultValue) {
            _listener.onPropertyStart(property.name);
            _inspect(value, property.schema);
            _listener.onPropertyEnd();
          }
        }
        
      } else {
        final mirror = reflect(object);
        for (int index = 0; index < length; index++) {
          final property = properties[index];
          final value = mirror.getField(property.getSymbol()).reflectee;
          
          if (! _omitDefaultValues || value != property.defaultValue) {
            _listener.onPropertyStart(property.name);
            _inspect(value, property.schema);
            _listener.onPropertyEnd();
          }
        }
      }
      
      _listener.onMapEnd();
      _popSeen();
      
    } else {
      assert(schema is IsList);
      assert(object is List && object is! TypedData);
      final listSchema = schema as IsList;
      final itemSchema = listSchema.schema;
      
      _pushSeen(object);
      _listener.onListStart();
      final length = object.length;
      for (var index = 0; index < length; index++) {
        _inspect(object[index], itemSchema);
        _listener.onListElement();
      }
      _listener.onListEnd();
      _popSeen();
    }
  }
  
  void encodeTo(object, CodecEncoder encoder, Schema schema,
                Buffer output, {bool omitDefaultValues: true}) {
    encoder.initialize(output);
    inspect(object, encoder, schema, omitDefaultValues: omitDefaultValues);
    encoder.clear();
  }
  
  List<int> encode(object, CodecEncoder encoder,
      Schema schema, {bool omitDefaultValues: true}) {
    globalBuffer.clear();
    encodeTo(object, encoder, schema,
        globalBuffer, omitDefaultValues: omitDefaultValues);
    final result = globalBuffer.bytes;
    globalBuffer.clear();
    return result;
  }
  
  String stringify(object, Schema schema) {
    final stringifier = new Stringifier();
    inspect(object, stringifier, schema);
    return stringifier.getResult();
  }
  
  void printObject(object, Schema schema) =>
    print(stringify(object, schema));
}
