library test.ormicida.orm.schema;

import 'dart:convert';
import 'package:unittest/unittest.dart';
import 'package:unittest/vm_config.dart';
import 'package:ormicida_json/ormicida_json.dart';
import 'package:ormicida_schema/ormicida_schema.dart';
import 'package:ormicida_schema_inspector/ormicida_schema_inspector.dart';

void main() {
  useVMConfiguration();
  runTests();
  
  final bytes = inspector.encode(dynamicPerson, json.encoder, null, omitDefaultValues: false);
  final string = UTF8.decode(bytes);
  print(string);
}



void runTests() {
  group('schema', () {
    group('IsDynamic', () {
      _testIsDynamic(const IsDynamic(true),
          null, name: 'nullable');
      _testIsDynamic(const IsDynamic(false), null);
      _testIsDynamic(const IsDynamic(), false);
      _testIsDynamic(const IsDynamic(), true);
      _testIsDynamic(const IsDynamic(), 0);
      _testIsDynamic(const IsDynamic(), 1.1);
      _testIsDynamic(const IsDynamic(), 'abc');
      _testIsDynamic(const IsDynamic(), []);
      _testIsDynamic(const IsDynamic(), {});
      _failIsDynamic(const IsBool(), null);
    });
  });
}

void _testIsDynamic(IsDynamic schema, value, {String name, Matcher matcher}) {
  test(name != null ? name : value.toString(), () {
    expect(schema.normalizeAndValidate(value), matcher != null ? matcher : equals(value));
  });
}

void _failIsDynamic(IsDynamic schema, value, {String name}) {
  if (name == null) name = value.toString();
  test(name, () {
    expect(() => schema.normalizeAndValidate(value), 
        throwsA(predicate((e) =>  e is SchemaException)));
  });
}

final dynamicPerson = {
  'id': 0,
  'name': 'Tibúrio Bombulá',
  'age': 45,
  'socialNumber': 24245233,
  'email': 'tiburcio@server.dom',
  'address1': {
    'street': 'Overthere 463',
    'code': '2345-678',
    'city': 'Overtheria'},
  'address2': {
    'street': 'Overthere 463',
    'code': '2345-678',
    'city': 'Overtheria'}
};

final mutablePerson = new MutPerson(
  id: 0,
  name: 'Tibúrio Bombulá',
  age: 45,
  socialNumber: 24245233,
  email: 'tiburcio@server.dom',
  address1: new MutAddress(
    street: 'Overthere 463',
    code: '2345-678',
    city: 'Overtheria'),
  address2: new MutAddress(
    street: 'Overthere 463',
    code: '2345-678',
    city: 'Overtheria')
);

final constantPerson = const ConstPerson(
  id: 0,
  name: 'Tibúrio Bombulá',
  age: 45,
  socialNumber: 24245233,
  email: 'tiburcio@server.dom',
  address1: const ConstAddress(
    street: 'Overthere 463',
    code: '2345-678',
    city: 'Overtheria'),
  address2: const ConstAddress(
    street: 'Overthere 463',
    code: '2345-678',
    city: 'Overtheria'));

const ADDRESS_PROPERTIES_STRING = const <Property>[
  const Property('street', const IsString()),
  const Property('code', const IsString()),
  const Property('city', const IsString()),
];
const ADDRESS_PROPERTIES_SYMBOL = const <Property>[
  const Property('street', const IsString(), symbol: const Symbol('street')),
  const Property('code', const IsString(), symbol: const Symbol('code')),
  const Property('city', const IsString(), symbol: const Symbol('city')),
];
final ADDRESS_SCHEMA_MUT_STRING =
  new IsObject(ADDRESS_PROPERTIES_STRING,
      constructor: (_) => new MutAddress());
final ADDRESS_SCHEMA_MUT_SYMBOL =
  new IsObject(ADDRESS_PROPERTIES_SYMBOL,
      constructor: (_) => new MutAddress());
final ADDRESS_SCHEMA_CONST_STRING =
  new IsObject(ADDRESS_PROPERTIES_STRING, isConst: true,
      constructor: (args) => new ConstAddress.fromJson(args));
final ADDRESS_SCHEMA_CONST_SYMBOL =
  new IsObject(ADDRESS_PROPERTIES_SYMBOL, isConst: true,
      constructor: (args) => new ConstAddress.fromJson(args));

final PERSON_SCHEMA_MUT_STRING = new IsObject(<Property>[
  const Property('id', const IsInt()),
  const Property('name', const IsString()),
  const Property('age', const IsInt()),
  const Property('socialNumber', const IsInt()),
  const Property('email', const IsString(), symbol: const Symbol('email')),
  new Property('address1', ADDRESS_SCHEMA_MUT_STRING),
  new Property('address2', ADDRESS_SCHEMA_MUT_STRING),
], constructor: (_) => new MutPerson());
final PERSON_SCHEMA_MUT_SYMBOL = new IsObject(<Property>[
  const Property('id', const IsInt(), symbol: const Symbol('id')),
  const Property('name', const IsString(), symbol: const Symbol('name')),
  const Property('age', const IsInt(), symbol: const Symbol('age')),
  const Property('socialNumber', const IsInt(), symbol: const Symbol('socialNumber')),
  const Property('email', const IsString(), symbol: const Symbol('email')),
  new Property('address1', ADDRESS_SCHEMA_MUT_SYMBOL, symbol: const Symbol('address1')),
  new Property('address2', ADDRESS_SCHEMA_MUT_SYMBOL, symbol: const Symbol('address2')),
], constructor: (_) => new MutPerson());
final PERSON_SCHEMA_CONST_STRING = new IsObject(<Property>[
  const Property('id', const IsInt()),
  const Property('name', const IsString()),
  const Property('age', const IsInt()),
  const Property('socialNumber', const IsInt()),
  const Property('email', const IsString()),
  new Property('address1', ADDRESS_SCHEMA_CONST_STRING),
  new Property('address2', ADDRESS_SCHEMA_CONST_STRING),
], isConst: true, constructor: (args) => new ConstPerson.fromJson(args));
final PERSON_SCHEMA_CONST_SYMBOL = new IsObject(<Property>[
  const Property('id', const IsInt(), symbol: const Symbol('id')),
  const Property('name', const IsString(), symbol: const Symbol('name')),
  const Property('age', const IsInt(), symbol: const Symbol('age')),
  const Property('socialNumber', const IsInt(), symbol: const Symbol('socialNumber')),
  const Property('email', const IsString(), symbol: const Symbol('email')),
  new Property('address1', ADDRESS_SCHEMA_CONST_SYMBOL, symbol: const Symbol('address1')),
  new Property('address2', ADDRESS_SCHEMA_CONST_SYMBOL, symbol: const Symbol('address2')),
], isConst: true, constructor: (params) => new ConstPerson.fromJson(params));

class MutAddress {
  String street;
  String code;
  String city;
  
  MutAddress({this.street, this.code, this.city});
  
  MutAddress.fromJson(Map json):
    street = json['street'],
    code = json['code'],
    city = json['city'];
  
  Map toJson() => {'street': street, 'code': code, 'city': city};
  
  operator ==(other) =>
      other is ConstAddress &&
      street == other.street &&
      code == other.code &&
      city == other.city;
  
  String toString() => 'MutAddress(street="$street", code="$code", city="$city")';
}

class MutPerson {
  int         id;
  String      name;
  int         age;
  int         socialNumber;
  String      email;
  MutAddress  address1;
  MutAddress  address2;
  
  MutPerson({
    this.id,
    this.name,
    this.age,
    this.socialNumber,
    this.email,
    this.address1,
    this.address2
  });
  
  MutPerson.fromJson(Map json):
    id = json['id'],
    name = json['name'],
    age = json['age'],
    socialNumber = json['socialNumber'],
    email = json['email'],
    address1 = json['address1'] != null ? new MutAddress.fromJson(json['address1']) : null,
    address2 = json['address2'] != null ? new MutAddress.fromJson(json['address2']) : null;
    
  Map toJson() => {
    'id': id,
    'name': name,
    'age': age,
    'socialNumber': socialNumber,
    'email': email,
    'address1': address1,
    'address2': address2
  };
  
  operator ==(other) =>
      other is MutPerson &&
      id == other.id &&
      name == other.name &&
      age == other.age &&
      socialNumber == other.socialNumber &&
      email == other.email &&
      address1 == other.address1 &&
      address2 == other.address2;
  
  String toString() => 'MutPerson(id=$id, name="$name", age=$age, socialNumber=$socialNumber, email="$email", address1=$address1, address2:$address2)';
}

class ConstAddress {
  final String street;
  final String code;
  final String city;
  
  const ConstAddress({this.street, this.code, this.city});
  
  factory ConstAddress.fromJson(Map json) =>
    new ConstAddress(
        street: json['street'],
        code: json['code'],
        city: json['city']);
  
  Map toJson() => {'street': street, 'code': code, 'city': city};
  
  operator ==(other) =>
      other is ConstAddress &&
      street == other.street &&
      code == other.code &&
      city == other.city;
  
  String toString() => 'ConstAddress(street="$street", code="$code", city="$city")';
}

class ConstPerson {
  final int           id;
  final String        name;
  final int           age;
  final int           socialNumber;
  final String        email;
  final ConstAddress  address1;
  final ConstAddress  address2;
  
  const ConstPerson({
    this.id,
    this.name,
    this.age,
    this.socialNumber,
    this.email,
    this.address1,
    this.address2
  });
  
  factory ConstPerson.fromJson(Map json) =>
    new ConstPerson(
      id: json['id'],
      name: json['name'],
      age: json['age'],
      socialNumber: json['socialNumber'],
      email: json['email'],
      address1: json['address1'] != null
                  ? new ConstAddress.fromJson(json['address1'])
                  : null,
      address2: json['address2'] != null
                  ? new ConstAddress.fromJson(json['address2'])
                  : null);
  
  Map toJson() => {
    'id': id,
    'name': name,
    'age': age,
    'socialNumber': socialNumber,
    'email': email,
    'address1': address1,
    'address2': address2
  };
  
  operator ==(other) =>
      other is ConstPerson &&
      id == other.id &&
      name == other.name &&
      age == other.age &&
      socialNumber == other.socialNumber &&
      email == other.email &&
      address1 == other.address1 &&
      address2 == other.address2;
  String toString() => 'ConstPerson(id=$id, name="$name", age=$age, socialNumber=$socialNumber, email="$email", address1=$address1, address2:$address2)';
}
